package me.hammerle.snuviengine.api;

import java.nio.ByteBuffer;
import me.hammerle.snuviengine.util.Color;
import static org.lwjgl.opengl.GL11.*;
import static org.lwjgl.opengl.GL15.*;
import static org.lwjgl.opengl.GL20.*;
import static org.lwjgl.opengl.GL30.*;

public final class FontRenderer {
    private static final char COLOR_CHAR = '#';
    private static final int FONT_HEIGHT = 8;
    private static final int FONT_WIDTH = 6;
    private static final int LINE_STEP = 1;
    private static final float SHADOW_STEP = 1.0f;
    private final static int BUFFER_BYTE_LENGTH = 4 * 1024 * 1024; // 4 MiB

    private final Texture[] FONT_TEXTURE = new Texture[]{
        new Texture("font8x8.png", true),
        new Texture("font16x16.png", true),
        new Texture("font24x24.png", true)
    };

    private final int vertexArray;
    private final int vertexBuffer;
    private int color = 0;
    private int scale = 1;
    private ByteBuffer buffer;
    private int previousOffset = 0;
    private int offset = BUFFER_BYTE_LENGTH;
    private int width = 0;
    private int height = 0;

    protected FontRenderer() {
        vertexArray = glGenVertexArrays();
        vertexBuffer = glGenBuffers();
        GLHelper.glBindVertexArray(vertexArray);
        GLHelper.glBindBuffer(vertexBuffer);
        glEnableVertexAttribArray(0);
        glEnableVertexAttribArray(1);
        glEnableVertexAttribArray(2);
        glVertexAttribPointer(0, 2, GL_FLOAT, false, 20, 0);
        glVertexAttribPointer(1, 2, GL_FLOAT, false, 20, 8);
        glVertexAttribPointer(2, 4, GL_UNSIGNED_BYTE, true, 20, 16);
    }

    protected void setScale(int scale) {
        this.scale = scale;
    }

    public float drawString(float x, float y, boolean shadow, String text) {
        buffer = getNextBuffer(120 * countPrintableCharacters(text) * (shadow ? 3 : 1));
        if(buffer == null) {
            return y;
        }
        if(shadow) {
            addStringToBuffer(x + SHADOW_STEP, y, true, text);
            addStringToBuffer(x + SHADOW_STEP, y + SHADOW_STEP, true, text);
        }
        y = addStringToBuffer(x, y, false, text);
        finishBuffer();
        drawBuffer();
        return y;
    }

    private int countPrintableCharacters(String text) {
        int count = 0;
        int index = 0;
        while(true) {
            index += consumeColor(text, index, false);
            if(index >= text.length()) {
                break;
            }
            if(text.charAt(index) != '\n') {
                count++;
            }
            index++;
        }
        return count;
    }

    public float drawString(float x, float y, String text) {
        return drawString(x, y, true, text);
    }

    private ByteBuffer getNextBuffer(int byteLength) {
        GLHelper.glBindBuffer(vertexBuffer);
        previousOffset = offset;
        offset += byteLength;
        if(offset + byteLength >= BUFFER_BYTE_LENGTH) {
            previousOffset = 0;
            offset = byteLength;
            glBufferData(GL_ARRAY_BUFFER, BUFFER_BYTE_LENGTH, GL_STREAM_DRAW);
        }
        return glMapBufferRange(GL_ARRAY_BUFFER, previousOffset, byteLength, GL_MAP_WRITE_BIT | GL_MAP_UNSYNCHRONIZED_BIT, buffer);
    }

    private float addStringToBuffer(float x, float y, boolean shadow, String text) {
        float oldX = x;
        setColor(0xFFFFFF, shadow);
        int index = 0;
        while(true) {
            index += consumeColor(text, index, shadow);
            if(index >= text.length()) {
                break;
            }
            char c = text.charAt(index);
            if(c == '\n') {
                y += FONT_HEIGHT + LINE_STEP;
                x = oldX;
            } else {
                addRectangle(x, y, c);
                x += FONT_WIDTH;
            }
            index++;
        }
        return y + FONT_HEIGHT + LINE_STEP;
    }

    private void setColor(int color, boolean shadow) {
        if(shadow) {
            this.color = Color.darken(color, 0.3f);
        } else {
            this.color = color;
        }
    }

    private int consumeColor(String text, int index, boolean shadow) {
        char c = getChar(text, index);
        if(c != COLOR_CHAR) {
            return 0;
        } else if(getChar(text, index + 1) == COLOR_CHAR) {
            return 1;
        } else if(text.length() < index + 7) {
            return 0;
        }
        setColor(parseColor(text, index + 1), shadow);
        return 7;
    }

    private char getChar(String text, int index) {
        return index < text.length() ? text.charAt(index) : '\0';
    }

    private int parseColor(String text, int index) {
        int c = 0;
        c += getCharValue(text.charAt(index)) << 4;
        c += getCharValue(text.charAt(index + 1));
        c += getCharValue(text.charAt(index + 2)) << 12;
        c += getCharValue(text.charAt(index + 3)) << 8;
        c += getCharValue(text.charAt(index + 4)) << 20;
        c += getCharValue(text.charAt(index + 5)) << 16;
        return c;
    }

    private int getCharValue(char c) {
        if(c >= '0' && c <= '9') {
            return c - '0';
        } else if(c >= 'A' && c <= 'F') {
            return c - 'A' + 10;
        } else if(c >= 'a' && c <= 'f') {
            return c - 'a' + 10;
        }
        return 0;
    }

    private void addRectangle(float minX, float minY, char c) {
        float tMinX = (c & 0xF) / 16.0f + 1.0f / 128.0f;
        float tMinY = (c >> 4) / 16.0f;
        float tMaxX = tMinX + 0.0625f - 2.0f / 128.0f;
        float tMaxY = tMinY + 0.0625f;
        float maxX = minX + FONT_WIDTH;
        float maxY = minY + FONT_HEIGHT;
        addToBuffer(minX, maxY, tMinX, tMaxY);
        addToBuffer(minX, minY, tMinX, tMinY);
        addToBuffer(maxX, maxY, tMaxX, tMaxY);
        addToBuffer(maxX, maxY, tMaxX, tMaxY);
        addToBuffer(minX, minY, tMinX, tMinY);
        addToBuffer(maxX, minY, tMaxX, tMinY);
    }

    private void addToBuffer(float x, float y, float tx, float ty) {
        buffer.putFloat(x);
        buffer.putFloat(y);
        buffer.putFloat(tx);
        buffer.putFloat(ty);
        buffer.putInt(color);
    }

    private void finishBuffer() {
        buffer.flip();
        glUnmapBuffer(GL_ARRAY_BUFFER);
    }

    private void drawBuffer() {
        GLHelper.glBindVertexArray(vertexArray);
        FONT_TEXTURE[Math.min(scale - 1, FONT_TEXTURE.length - 1)].bind();
        glDrawArrays(GL_TRIANGLES, previousOffset / 20, buffer.limit() / 20);
    }

    public FontRenderer getSize(String text) {
        width = 0;
        height = FONT_HEIGHT + LINE_STEP;
        int currentWidth = 0;
        int index = 0;
        while(true) {
            index += consumeColor(text, index, false);
            if(index >= text.length()) {
                break;
            }
            char c = text.charAt(index);
            if(c == '\n') {
                width = Math.max(currentWidth, width);
                currentWidth = 0;
                height += FONT_HEIGHT + LINE_STEP;
            } else {
                currentWidth++;
            }
            index++;
        }
        width = Math.max(currentWidth, width) * FONT_WIDTH;
        return this;
    }

    public int getStringWidth() {
        return width;
    }

    public int getStringHeight() {
        return height;
    }

    public float getCharWidth() {
        return FONT_WIDTH;
    }

    public float getCharHeight() {
        return FONT_HEIGHT + LINE_STEP;
    }
}
